mod category;
mod consts;
mod crate_info;
mod scraper;
mod search_item;
mod version;

use scraper::{get_home_page, search_by_criteria, search_crate};
use tera::Tera;
use tokio::sync::OnceCell;

static TERA: OnceCell<Tera> = OnceCell::const_new();

async fn render<S: AsRef<str>>(name: S, context: &tera::Context) -> Result<String, ()> {
    let tera = TERA
        .get_or_init(|| async {
            let mut tera = Tera::default();
            tera.register_filter("separate_with_commas", tera_thousands::separate_with_commas);
            for path in glob::glob("templates/*").unwrap().flatten() {
                let raw_path = path.clone();
                let filename = raw_path
                    .file_name()
                    .unwrap() // This should be safe (?
                    .to_str()
                    .unwrap() // ._.?
                    .split('.')
                    .next();
                tera.add_template_file(path, filename).unwrap()
            }
            tera
        })
        .await;
    tera.render(name.as_ref(), context).map_err(|_| ())
}
#[windmark::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    windmark::router::Router::new()
        .set_private_key_file("key.pem")
        .set_certificate_file("cert.pem")
        .set_port(1971)
        .enable_default_logger(true)
        .set_fix_path(true)
        .mount("/", move |_| async {
            let mut context = tera::Context::new();
            let (total_crates_indexed, categories) = get_home_page();

            context.insert("categories", &categories);
            context.insert("total_crates_indexed", &total_crates_indexed);
            windmark::response::Response::success(render("index", &context).await.unwrap())
        })
        .mount("/about", move |_| async {
            let context = tera::Context::new();

            windmark::response::Response::success(render("about", &context).await.unwrap())
        })
        .mount("/search/:criteria", move |request| async move {
            let mut context = tera::Context::new();

            let criteria = urlencoding::decode(request.parameters.get("criteria").unwrap())
                .unwrap()
                .to_string();

            let (error, items) = match search_by_criteria(criteria.to_string()) {
                None => (Some("Nothing found :("), None),
                Some(items) => (None, Some(items)),
            };

            context.insert("criteria", &criteria);
            context.insert("error", &error);
            context.insert("items", &items);
            windmark::response::Response::success(render("search", &context).await.unwrap())
        })
        .mount("/crates/:crate_name", move |request| async move {
            let mut context = tera::Context::new();

            let crate_name = request.parameters.get("crate_name").unwrap();

            let (error, crate_info) = match search_crate(crate_name.to_string()) {
                None => (Some("Crate not found"), None),
                Some(crate_info) => (None, Some(crate_info)),
            };
            context.insert("error", &error);
            context.insert("item", &crate_info);
            windmark::response::Response::success(render("crate", &context).await.unwrap())
        })
        .run()
        .await
}
